@@ -21,9 +21,11 @@
 
 #include <bzlib.h>
 #include <zlib.h>
+#include <lzma.h>
 
 #include "util.h"
 #include "md5.h"
+#include "sha256.h"
 #include "rpmhead.h"
 #include "cpio.h"
 #include "cfile.h"
@@ -394,10 +396,7 @@ fillblock_prelink(struct blk *b, int id, struct seqdescr *sd, struct fileblock *
 	    {
 	      if (fd != -1)
 		{
-		  if (isp)
-		    prelinked_close(fd);
-		  else
-		    close(fd);
+		  close(fd);
 		  fd = -1;
 		}
 	      sd++;
@@ -486,10 +485,7 @@ fillblock_prelink(struct blk *b, int id, struct seqdescr *sd, struct fileblock *
 		{
 		  if (fd)
 		    {
-		      if (isp)
-			prelinked_close(fd);
-		      else
-			close(fd);
+		      close(fd);
 		      fd = -1;
 		    }
 		  l2 = l > sd->datalen - o ? sd->datalen - o : l;
@@ -936,19 +932,6 @@ getblock(int id, struct seqdescr *sdesc, int nsdesc, struct fileblock *fb, int i
   return b;
 }
 
-char *comp2str(int comp)
-{
-  if (comp == CFILE_COMP_BZ)
-    return "bzip2";
-  if (comp == CFILE_COMP_GZ)
-    return "gzip";
-  if (comp == CFILE_COMP_GZ_RSYNC)
-    return "gzip rsyncable";
-  if (comp == CFILE_COMP_UN)
-    return "uncompressed";
-  return "???";
-}
-
 int
 cfile_write_uncomp(struct cfile *f, void *buf, int len)
 {
@@ -974,13 +957,63 @@ cfile_write_uncomp(struct cfile *f, void *buf, int len)
   return len;
 }
 
+typedef struct {
+  union {
+    MD5_CTX md5ctx;
+    SHA256_ctx sha256ctx;
+  } ctx;
+} DIG_CTX;
+
+
+static inline void
+DIG_Init(DIG_CTX *ctx, int digestalgo)
+{
+  if (digestalgo == 1)
+    rpmMD5Init(&ctx->ctx.md5ctx);
+  else if (digestalgo == 8)
+    SHA256_init(&ctx->ctx.sha256ctx);
+}
+
+static inline void
+DIG_Update(DIG_CTX *ctx, int digestalgo, unsigned char const *buf, unsigned len)
+{
+  if (digestalgo == 1)
+    rpmMD5Update(&ctx->ctx.md5ctx, buf, len);
+  else if (digestalgo == 8)
+    SHA256_update(&ctx->ctx.sha256ctx, buf, len);
+}
+
+static inline void
+DIG_Final(DIG_CTX *ctx, int digestalgo, unsigned char *digest)
+{
+  if (digestalgo == 1)
+    rpmMD5Final(digest, &ctx->ctx.md5ctx);
+  else if (digestalgo == 8)
+    {
+      SHA256_final(&ctx->ctx.sha256ctx);
+      SHA256_digest(&ctx->ctx.sha256ctx, digest);
+    }
+  else
+    *digest = 0;
+}
+
+static inline int
+DIG_Len(int digestalgo)
+{
+  if (digestalgo == 1)
+    return 16;
+  if (digestalgo == 8)
+    return 32;
+  return 0;
+}
+
 int
-checkprelinked(char *name, unsigned char *hmd5, unsigned int size)
+checkprelinked(char *name, int digestalgo, unsigned char *hmd5, unsigned int size)
 {
   int fd, l;
   unsigned char buf[4096];
-  MD5_CTX ctx;
-  unsigned char md5[16];
+  DIG_CTX ctx;
+  unsigned char md5[32];
 
   nprelink++;
   if ((fd = prelinked_open(name)) < 0)
@@ -988,17 +1021,17 @@ checkprelinked(char *name, unsigned char *hmd5, unsigned int size)
       perror(name);
       return -1;
     }
-  rpmMD5Init(&ctx);
+  DIG_Init(&ctx, digestalgo);
   while (size && (l = read(fd, buf, sizeof(buf))) > 0)
     {
       if (l > size)
 	l = size;
-      rpmMD5Update(&ctx, buf, l);
+      DIG_Update(&ctx, digestalgo, buf, l);
       size -= l;
     }
-  prelinked_close(fd);
-  rpmMD5Final(md5, &ctx);
-  if (memcmp(md5, hmd5, 16))
+  close(fd);
+  DIG_Final(&ctx, digestalgo, md5);
+  if (memcmp(md5, hmd5, DIG_Len(digestalgo)))
     {
       fprintf(stderr, "%s: contents have been changed\n", name);
       return -1;
@@ -1007,12 +1040,12 @@ checkprelinked(char *name, unsigned char *hmd5, unsigned int size)
 }
 
 int
-checkfilemd5(char *name, unsigned char *hmd5, unsigned int size)
+checkfilemd5(char *name, int digestalgo, unsigned char *hmd5, unsigned int size)
 {
   int fd, l;
   unsigned char buf[4096];
-  MD5_CTX ctx;
-  unsigned char md5[16];
+  DIG_CTX ctx;
+  unsigned char md5[32];
   struct stat stb;
 
   if ((fd = open(name, O_RDONLY)) < 0 || fstat(fd, &stb))
@@ -1020,29 +1053,29 @@ checkfilemd5(char *name, unsigned char *hmd5, unsigned int size)
       perror(name);
       return -1;
     }
-  rpmMD5Init(&ctx);
+  DIG_Init(&ctx, digestalgo);
   if (stb.st_size > size && (l = read(fd, buf, sizeof(buf))) > 0)
     {
       if (is_prelinked(fd, buf, l))
 	{
 	  close(fd);
-	  return checkprelinked(name, hmd5, size);
+	  return checkprelinked(name, digestalgo, hmd5, size);
 	}
       if (l > size)
 	l = size;
-      rpmMD5Update(&ctx, buf, l);
+      DIG_Update(&ctx, digestalgo, buf, l);
       size -= l;
     }
   while (size && (l = read(fd, buf, sizeof(buf))) > 0)
     {
       if (l > size)
 	l = size;
-      rpmMD5Update(&ctx, buf, l);
+      DIG_Update(&ctx, digestalgo, buf, l);
       size -= l;
     }
   close(fd);
-  rpmMD5Final(md5, &ctx);
-  if (memcmp(md5, hmd5, 16))
+  DIG_Final(&ctx, digestalgo, md5);
+  if (memcmp(md5, hmd5, DIG_Len(digestalgo)))
     {
       fprintf(stderr, "%s: contents have been changed\n", name);
       return -1;
@@ -1051,7 +1084,7 @@ checkfilemd5(char *name, unsigned char *hmd5, unsigned int size)
 }
 
 int
-checkfilesize(char *name, unsigned char *hmd5, unsigned int size)
+checkfilesize(char *name, int digestalgo, unsigned char *hmd5, unsigned int size)
 {
   struct stat stb;
   unsigned char buf[128];
@@ -1071,7 +1104,7 @@ checkfilesize(char *name, unsigned char *hmd5, unsigned int size)
       if (fd != -1 && (l = read(fd, buf, sizeof(buf))) > 0 && is_prelinked(fd, buf, l))
 	{
 	  close(fd);
-	  return checkprelinked(name, hmd5, size);
+	  return checkprelinked(name, digestalgo, hmd5, size);
 	}
       if (fd != -1)
         close(fd);
@@ -1110,6 +1143,7 @@ main(int argc, char **argv)
   unsigned char buf[4096];
   MD5_CTX wrmd5;
   unsigned char wrmd5res[16];
+  int nofullmd5 = 0;
   FILE *ofp;
   int numblks;
   int percent = 0;
@@ -1128,8 +1162,9 @@ main(int argc, char **argv)
   int seqmatches = 1;
   FILE *vfp;
   struct deltarpm d;
+  char *arch = 0;
 
-  while ((c = getopt(argc, argv, "cCisvpr:")) != -1)
+  while ((c = getopt(argc, argv, "cCisvpr:a:")) != -1)
     {
       switch(c)
 	{
@@ -1157,6 +1192,9 @@ main(int argc, char **argv)
 	  checkflags = SEQCHECK_SIZE;
 	  check = 1;
 	  break;
+	case 'a':
+	  arch = optarg;
+	  break;
 	default:
 	  fprintf(stderr, "usage: applydeltarpm [-r <rpm>] deltarpm rpm\n");
           exit(1);
@@ -1214,6 +1252,7 @@ main(int argc, char **argv)
       if (verbose)
 	fprintf(vfp, "reading deltarpm\n");
       readdeltarpm(deltarpm, &d, &bfp);
+      nofullmd5 = !memcmp("\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", d.targetmd5, 16);
 #ifdef DELTARPM_64BIT
       if (d.outlen >= 0xffffffffULL << BLKSHIFT)
 	{
@@ -1252,13 +1291,17 @@ main(int argc, char **argv)
   addblkcomp = CFILE_COMP_BZ;
   if (d.addblklen > 9 && d.addblk[0] == 0x1f && d.addblk[1] == 0x8b)
     addblkcomp = CFILE_COMP_GZ;
+  else if (d.addblklen > 3 && (d.addblk[0] == 255 && d.addblk[1] == 'L' && d.addblk[2] == 'Z'))
+    addblkcomp = CFILE_COMP_LZMA;
+  else if (d.addblklen > 6 && (d.addblk[0] == 0xfd && d.addblk[1] == '7' && d.addblk[2] == 'z' && d.addblk[3] == 'X' && d.addblk[4] == 'Z'))
+    addblkcomp = CFILE_COMP_XZ;
   if (info)
     {
       unsigned int *size;
       if (d.version)
 	printf("deltarpm version: %c\n", d.version & 0xff);
       printf("deltarpm type: %s\n", d.h ? "standard" : d.targetcomp != CFILE_COMP_UN || d.inn != 0 || d.outn != 0 ? "rpm-only" : "rpm-only, no diff");
-      printf("deltarpm compression: %s\n", comp2str(d.deltacomp));
+      printf("deltarpm compression: %s\n", cfile_comp2str(d.deltacomp));
       printf("sequence: %s-", d.nevr);
       for (i = 0; i < d.seql; i++)
 	printf("%02x", d.seq[i]);
@@ -1271,7 +1314,7 @@ main(int argc, char **argv)
 	{
 	  printf("target payload size: %llu\n", (unsigned long long)d.paylen);
 	  if (d.targetcomp != CFILE_COMP_XX)
-	    printf("target payload compression: %s\n", comp2str(d.targetcomp));
+	    printf("target payload compression: %s\n", cfile_comp2str(d.targetcomp));
 	}
       if (d.targetsize == 0)
 	{
@@ -1294,7 +1337,7 @@ main(int argc, char **argv)
 	  printf("internal data size: %llu\n", (unsigned long long)d.inlen);
 	  printf("compressed add data size: %d\n", d.addblklen);
 	  if (d.addblklen)
-	    printf("compressed add data compression: %s\n", comp2str(addblkcomp));
+	    printf("compressed add data compression: %s\n", cfile_comp2str(addblkcomp));
 	  printf("instructions: %d\n", d.inn + d.outn);
 	}
       if (bfp)
@@ -1307,7 +1350,7 @@ main(int argc, char **argv)
       fprintf(stderr, "deltarpm contains unknown compression parameters\n");
       exit(1);
     }
-  
+
   if (!fromrpm)
     {
       pid_t pid;
@@ -1341,7 +1384,10 @@ main(int argc, char **argv)
 	      dup2(pi[1], 1);
 	      close(pi[1]);
 	    }
-	  execlp(RPMDUMPHEADER, RPMDUMPHEADER, d.nevr, (char *)0);
+	  if (arch)
+	    execlp(RPMDUMPHEADER, RPMDUMPHEADER, "-a", arch, d.nevr, (char *)0);
+	  else
+	    execlp(RPMDUMPHEADER, RPMDUMPHEADER, d.nevr, (char *)0);
 	  perror(RPMDUMPHEADER);
 	  _exit(1);
 	}
@@ -1410,7 +1456,7 @@ main(int argc, char **argv)
     }
   if (d.h || seqcheck)
     {
-      int (*checkfunc)(char *, unsigned char *, unsigned int);
+      int (*checkfunc)(char *, int, unsigned char *, unsigned int);
       if (headtofb(h, &fb))
 	{
 	  fprintf(stderr, "bad header\n");
@@ -1464,7 +1510,8 @@ main(int argc, char **argv)
       fprintf(stderr, "write error\n");
       exit(1);
     }
-  rpmMD5Update(&wrmd5, d.lead, d.leadl);
+  if (!nofullmd5)
+    rpmMD5Update(&wrmd5, d.lead, d.leadl);
   if (!d.h)
     fromrpm_raw = 1;
 
@@ -1493,7 +1540,24 @@ main(int argc, char **argv)
 	  exit(1);
 	}
       rpmMD5Final(wrmd5res, &wrmd5);
-      if (memcmp(wrmd5res, d.targetmd5, 16) != 0)
+      if (nofullmd5)
+        {
+          struct rpmhead *dsigh = readhead_buf(d.lead + 96, d.leadl - 96, 0);
+          if (dsigh)
+            {
+              unsigned char *hmd5 = headbin(dsigh, SIGTAG_MD5, 16);
+              if (hmd5)
+                {
+                  if (memcmp(wrmd5res, hmd5, 16) != 0)
+                    {
+                      fprintf(stderr, "%s: md5 mismatch of result\n", deltarpm);
+                      exit(1);
+                    }
+                }
+              xfree(dsigh);
+            }
+        }
+      else if (memcmp(wrmd5res, d.targetmd5, 16) != 0)
 	{
 	  fprintf(stderr, "%s: md5 mismatch of result\n", deltarpm);
 	  exit(1);
@@ -1752,7 +1816,24 @@ main(int argc, char **argv)
         fprintf(vfp, "had to call prelink %d times\n", nprelink);
     }
   rpmMD5Final(wrmd5res, &wrmd5);
-  if (memcmp(wrmd5res, d.targetmd5, 16) != 0)
+  if (nofullmd5)
+    {
+      struct rpmhead *dsigh = readhead_buf(d.lead + 96, d.leadl - 96, 0);
+      if (dsigh)
+        {
+          unsigned char *hmd5 = headbin(dsigh, SIGTAG_MD5, 16);
+          if (hmd5)
+            {
+              if (memcmp(wrmd5res, hmd5, 16) != 0)
+                {
+                  fprintf(stderr, "%s: md5 mismatch of result\n", deltarpm);
+                  exit(1);
+                }
+            }
+          xfree(dsigh);
+        }
+    }
+  else if (memcmp(wrmd5res, d.targetmd5, 16) != 0)
     {
       fprintf(stderr, "%s: md5 mismatch of result\n", deltarpm);
       exit(1);
