@@ -5,6 +5,10 @@
  * for further information
  */
 
+#include <zlib.h>
+#include <bzlib.h>
+#include <lzma.h>
+
 struct cfile {
   int fd;
   void *fp;
