@@ -219,10 +219,209 @@ is_unpatched(char *n, struct rpmlfile *files1, int nfiles1, struct rpmlfile *fil
   return 0;
 }
 
+struct streamdata {
+  bsuint bsize;
+
+  unsigned char *old;
+  bsuint oldl;
+  bsuint oldskip;
+  int oldeof;
+
+  unsigned char *new;
+  bsuint newl;
+  bsuint newskip;
+  int neweof;
+
+  struct cfile *newf;
+  unsigned char *xnewdata;
+  bsuint xnewdatal;
+
+  void *stepd;
+  bsuint scan;
+  bsuint lastpos;
+  bsuint lastscan;
+
+  struct cfile *cfa;	/* add block */
+  struct cfile *cfi;	/* in data */
+
+  struct instr *instr;
+  int instrlen;
+};
+
+struct streamdata sd;	/* hack: global for now */
+
+void
+addtocpio_stream(unsigned char *d, int l)
+{
+  bsuint bsize = sd.bsize;
+  if (sd.lastscan == sd.newl && sd.neweof)
+    {
+      sd.oldskip += sd.oldl + l;
+      sd.oldl = 0;
+      return;
+    }
+  for (;;)
+    {
+      if (sd.oldl < bsize && !sd.oldeof)
+	{
+	  int l2 = bsize - sd.oldl;
+	  if (l2 > l)
+	    l2 = l;
+	  memcpy(sd.old + sd.oldl, d, l2);
+	  sd.oldl += l2;
+          l -= l2;
+	  d += l2;
+	  if (sd.oldl < bsize)
+	    return;
+	}
+      while (sd.newl < bsize && !sd.neweof)
+	{
+	  int l2 = bsize - sd.newl;
+	  if (sd.xnewdatal)
+	    {
+	      if (l2 > sd.xnewdatal)
+		l2 = sd.xnewdatal;
+	      memcpy(sd.new + sd.newl, sd.xnewdata, l2);
+	      sd.newl += l2;
+	      sd.xnewdata += l2;
+	      sd.xnewdatal -= l2;
+	      continue;
+	    }
+	  l2 = sd.newf->read(sd.newf, sd.new + sd.newl, l2);
+	  if (l2 < 0)
+	    {
+	      fprintf(stderr, "payload read failed\n");
+	      exit(1);
+	    }
+	  if (l2 == 0)
+	    sd.neweof = 1;
+	  sd.newl += l2;
+	}
+      if (sd.lastscan != sd.newl)
+	{
+	  struct instr instr;
+	  mkdiff_step(sd.stepd, sd.old, sd.oldl, sd.new, sd.newl, &instr, &sd.scan, &sd.lastpos, &sd.lastscan);
+	  if (instr.copyout && !sd.oldeof && (sd.lastscan == sd.newl || instr.copyoutoff + instr.copyout == sd.oldl))
+	    {
+	      /* incomplete match, ignore indata part */
+	      instr.copyin = 0;
+	      sd.scan = sd.lastscan = instr.copyinoff;
+	      sd.lastpos = instr.copyoutoff + instr.copyout;
+	    }
+	  else if (!instr.copyout && sd.lastscan == sd.newl)
+	    {
+	      /* no match found in old data, advance if we're behind */
+	      if (sd.lastpos + sd.oldskip < sd.lastscan + sd.newskip)
+	        sd.lastpos = sd.oldl;
+	    }
+	  /* printf("INSTR: %d@%d %d@%d\n", instr.copyout, instr.copyoutoff + sd.oldskip, instr.copyin, instr.copyinoff + sd.newskip); */
+	  if (instr.copyin)
+	    {
+	      if (instr.copyinoff + instr.copyin > sd.newl)
+		abort();
+	      if (sd.cfi->write(sd.cfi, sd.new + instr.copyinoff, instr.copyin) != instr.copyin)
+		{
+		  fprintf(stderr, "could not create indata block\n");
+		  exit(1);
+		}
+	    }
+	  if (instr.copyout && sd.cfa)
+	    {
+	      bsuint lenf = instr.copyout;
+	      bsuint lastpos = instr.copyoutoff;
+	      bsuint lastscan = instr.copyinoff - lenf;
+	      if (instr.copyoutoff + instr.copyout > sd.oldl)
+		abort();
+	      if (instr.copyinoff - lenf + instr.copyout > sd.newl)
+		abort();
+	      while (lenf > 0) 
+		{
+		  unsigned char addblk[4096];
+		  int len2, i;
+
+		  len2 = lenf > 4096 ? 4096 : lenf;
+		  for (i = 0; i < len2; i++) 
+		    addblk[i] = sd.new[lastscan + i] - sd.old[lastpos + i];
+		  if (sd.cfa->write(sd.cfa, addblk, len2) != len2)
+		    {
+		      fprintf(stderr, "could not create compressed add block\n");
+		      exit(1);
+		    }
+		  lastscan += len2;
+		  lastpos += len2;
+		  lenf -= len2;
+		}
+	    }
+	  instr.copyinoff += sd.newskip;
+	  instr.copyoutoff += sd.oldskip;
+	  if (sd.instrlen && sd.instr[sd.instrlen - 1].copyin == 0 && sd.instr[sd.instrlen - 1].copyoutoff + sd.instr[sd.instrlen - 1].copyout == instr.copyoutoff)
+	    {
+	      /* just add to last instruction */
+	      sd.instr[sd.instrlen - 1].copyin = instr.copyin;
+	      sd.instr[sd.instrlen - 1].copyinoff = instr.copyinoff;
+	      sd.instr[sd.instrlen - 1].copyout += instr.copyout;
+	    }
+	  else if (instr.copyin || instr.copyout)
+	    {
+	      if ((sd.instrlen & 31) == 0)
+		{
+		  if (sd.instr)
+		    sd.instr = realloc(sd.instr, sizeof(*sd.instr) * (sd.instrlen + 32));
+		  else
+		    sd.instr = malloc(sizeof(*sd.instr) * (sd.instrlen + 32));
+		  if (!sd.instr)
+		    {
+		      fprintf(stderr, "out of memory\n");
+		      exit(1);
+		    }
+		}
+	      sd.instr[sd.instrlen] = instr;
+	      sd.instrlen++;
+	    }
+	}
+      if (sd.lastscan > bsize / 4 && !sd.neweof)
+	{
+	  if (sd.newl > sd.lastscan)
+	    memmove(sd.new, sd.new + sd.lastscan, sd.newl - sd.lastscan);
+	  sd.newl -= sd.lastscan;
+	  sd.newskip += sd.lastscan;
+	  sd.scan -= sd.lastscan;
+	  sd.lastscan = 0;
+	}
+      if (sd.lastpos > bsize / 2 && !sd.oldeof)
+	{
+	  bsuint move = sd.lastpos - bsize / 4;
+	  if (sd.lastpos + sd.oldskip + bsize < sd.lastscan + sd.newskip)
+	    move = sd.lastpos;
+	  if (move > sd.lastpos)
+	    move = sd.lastpos;
+	  if (sd.oldl > move)
+	    memmove(sd.old, sd.old + move, sd.oldl - move);
+	  sd.oldl -= move;
+	  sd.oldskip += move;
+	  sd.lastpos -= move;
+	  mkdiff_step_freedata(sd.stepd);
+	}
+      if (sd.lastscan == sd.newl && sd.neweof)
+	{
+	  sd.oldskip += sd.oldl + l;
+	  sd.oldl = 0;
+	  mkdiff_step_freedata(sd.stepd);
+	  return;
+	}
+    }
+}
+
 void
 addtocpio(unsigned char **cpiop, bsuint *cpiol, unsigned char *d, int l)
 {
   bsuint cpl = *cpiol;
+  if (*cpiop == (unsigned char *)&sd)
+    {
+      *cpiol += l;
+      addtocpio_stream(d, l);
+      return;
+    }
   if (cpl + l < cpl || cpl + l + 65535 < cpl + l)
     {
       fprintf(stderr, "cpio archive to big\n");
@@ -234,22 +433,19 @@ addtocpio(unsigned char **cpiop, bsuint *cpiol, unsigned char *d, int l)
   *cpiol = cpl + l;
 }
 
-unsigned char **
-convertinstr(struct instr *instr, int instrlen, struct deltarpm *d, unsigned char *new)
+void
+convertinstr(struct instr *instr, int instrlen, struct deltarpm *d)
 {
   int i, j;
   bsuint x1, x2, x3, off;
-  unsigned int insize;
   unsigned int *b1;
   int nb1;
   unsigned int *b2;
   int nb2;
-  unsigned char **indatalist;
 
   b1 = b2 = 0;
   nb1 = nb2 = 0;
   j = 0;
-  insize = 0;
   off = 0;
   for (i = 0; i < instrlen; i++)
     {
@@ -307,13 +503,11 @@ retry2:
 	    {
 	      b1[2 * nb1 + 1] = 0x7fffffff;
 	      x2 -= 0x7fffffff;
-              insize += 0x7fffffff;
 	      nb1++;
 	      goto retry2;
 	    }
           b1[2 * nb1 + 1] = x2;
           nb1++;
-          insize += x2;
         }
     }
   if (j)
@@ -328,18 +522,47 @@ retry2:
   d->in = b1;
   d->outn = nb2;
   d->out = b2;
-  indatalist = xmalloc2(nb1, sizeof(unsigned char *));
-  for (i = j = 0; i < instrlen; i++)
-    if (instr[i].copyin)
-      indatalist[j++] = new + instr[i].copyinoff;
-  d->inlen = insize;
-  d->indata = 0;
+}
+
+unsigned char **
+createindatalist(struct instr *instr, int instrlen, struct deltarpm *d, unsigned char *new)
+{
+  int i, j;
+  bsuint off, left, todo;
+  unsigned char **indatalist;
+  
+  left = off = 0;
+  j = 0;
+  d->inlen = 0;
+  indatalist = xcalloc(d->inn, sizeof(unsigned char *));
+  for (i = 0; i < d->inn; i++)
+    {
+      todo = d->in[2 * i + 1];
+      while (!left && todo)
+	{
+	  off = instr[j].copyinoff;
+	  left = instr[j].copyin;
+	  j++;
+	}
+      indatalist[i] = new + off;
+      off += todo;
+      left -= todo;
+      d->inlen += todo;
+    }
   return indatalist;
 }
 
 int
 str2comp(char *comp)
 {
+  int n = strlen(comp);
+  if (n > 2 && n < 20 && comp[n - 2] == '.' && comp[n - 1] >= '0' && comp[n - 1] <= '9')
+    {
+      char buf[20];
+      strcpy(buf, comp);
+      buf[n - 2] = 0;
+      return cfile_setlevel(str2comp(buf), comp[n - 1] - '0');
+    }
   if (!strcmp(comp, "bzip2"))
     return CFILE_COMP_BZ;
   if (!strcmp(comp, "gzip"))
@@ -442,7 +665,8 @@ main(int argc, char **argv)
   unsigned int *fileflags, *filemodes, *filerdevs, *filesizes, *fileverify, *filecolors;
   int digestalgo = 1;
   unsigned int *digestalgoarray;
-  int i, fd, l, l2, l3;
+  int i, l, l2, l3;
+  int fd, nfd;
   struct cfile *bfd;
   struct cpiophys cph;
   char *namebuf;
@@ -494,7 +718,7 @@ main(int argc, char **argv)
   int verbose = 0;
   int version = 3;
   struct deltarpm d;
-  unsigned char **indatalist;
+  unsigned char **indatalist = 0;
   int rpmonly = 0;
   int alone = 0;
   FILE *vfp = 0;
@@ -505,8 +729,11 @@ main(int argc, char **argv)
   int targetcomp = CFILE_COMP_XX;
   char *payloadflags;
 
+  bsuint stream = 0;
+
   memset(&d, 0, sizeof(d));
-  while ((c = getopt(argc, argv, "vV:prl:s:z:u")) != -1)
+  memset(&sd, 0, sizeof(sd));
+  while ((c = getopt(argc, argv, "vV:prl:s:z:um:")) != -1)
     {
       switch (c)
 	{
@@ -534,6 +761,9 @@ main(int argc, char **argv)
 	case 'z':
 	  compopt = optarg;
 	  break;
+	case 'm':
+	  stream = atoi(optarg) * (1024 * 256);
+	  break;
 	default:
 	  fprintf(stderr, "usage: makedeltarpm [-l <file>] [-s seq] oldrpm newrpm deltarpm\n");
 	  exit(1);
@@ -631,7 +861,12 @@ main(int argc, char **argv)
 
   rpmMD5Init(&seqmd5);
 
+  /* open old rpm */
+  /* (if alone == 1,  oldrpm == newrpm) */
   rpmname = argv[argc - 3 + alone];
+  if (verbose >= 2)
+    fprintf(vfp, "Old: %s\n", argv[argc - 3 + alone]);
+
   if (!strcmp(rpmname, "-"))
     fd = 0;
   else if ((fd = open(rpmname, O_RDONLY)) < 0)
@@ -665,6 +900,7 @@ main(int argc, char **argv)
 
   if (alone && rpmonly)
     {
+      /* this mode just updates the lead and signatures, no need to do a real diff */
       if (verbose)
 	fprintf(vfp, "reading rpm header...\n");
       rpmMD5Init(&fullmd5);
@@ -733,10 +969,13 @@ main(int argc, char **argv)
       d.lead = xfree(d.lead);
       d.leadl = 0;
       seq = xfree(seq);
+      d.targetnevr = xfree(d.targetnevr);
       exit(0);
     }
+
   if (!alone)
     sigh = xfree(sigh);
+
   if (pinfo)
     {
       if (strcmp(nevr, nevr1) != 0)
@@ -771,37 +1010,108 @@ main(int argc, char **argv)
       exit(1);
     }
 
+/***************************************************************************/
+
   if (alone)
     {
-      d.h = h;
       if (verbose)
 	fprintf(vfp, "reading rpm...\n");
-      rpmMD5Init(&fullmd5);
-      rpmMD5Update(&fullmd5, rpmlead, 96);
-      rpmMD5Update(&fullmd5, sigh->intro, 16);
-      rpmMD5Update(&fullmd5, sigh->data, sigh->cnt * 16 + sigh->dcnt);
-      rpmMD5Update(&fullmd5, d.h->intro, 16);
-      rpmMD5Update(&fullmd5, d.h->data, d.h->cnt * 16 + d.h->dcnt);
-      fullsize = 96 + 16 + sigh->cnt * 16 + sigh->dcnt + 16 + d.h->cnt * 16 + d.h->dcnt;
-      newbz = cfile_open(CFILE_OPEN_RD, fd, 0, CFILE_COMP_XX, CFILE_LEN_UNLIMITED, (cfile_ctxup)rpmMD5Update, &fullmd5);
-      if (!newbz)
+      nfd = fd;
+      fd = -1;
+      d.h = h;
+      h = 0;
+      stream = 0;	/* sorry! */
+    }
+  else
+    {
+      if (verbose >= 2)
+	fprintf(vfp, "New: %s\n", argv[argc - 2]);
+
+      if (verbose)
+	fprintf(vfp, "reading new rpm...\n");
+ 
+      rpmname = argv[argc - 2];
+      if (!strcmp(rpmname, "-"))
+	nfd = 0;
+      else if ((nfd = open(rpmname, O_RDONLY)) < 0)
 	{
-	  fprintf(stderr, "payload open failed\n");
+	  perror(rpmname);
 	  exit(1);
 	}
-      if (cfile_detect_rsync(newbz))
+      if (read(nfd, rpmlead, 96) != 96 || rpmlead[0] != 0xed || rpmlead[1] != 0xab || rpmlead[2] != 0xee || rpmlead[3] != 0xdb)
 	{
-	  fprintf(stderr, "detect_rsync failed\n");
+	  fprintf(stderr, "%s: not a rpm\n", rpmname);
 	  exit(1);
 	}
-      targetcomp = newbz->comp;
-      if ((payloadflags = headstring(d.h, TAG_PAYLOADFLAGS)) != 0)
-	if (*payloadflags >= '1' && *payloadflags <= '9')
-	  targetcomp = cfile_setlevel(targetcomp, *payloadflags - '0');
-      if (paycomp == CFILE_COMP_XX)
-	paycomp = targetcomp;
-      if (addblkcomp == CFILE_COMP_XX)
-	addblkcomp = targetcomp;
+      if (rpmlead[4] != 0x03 || rpmlead[0x4e] != 0 || rpmlead[0x4f] != 5)
+	{
+	  fprintf(stderr, "%s: not a v3 rpm or not new header styles\n", rpmname);
+	  exit(1);
+	}
+      sigh = readhead(nfd, 1);
+      if (!sigh)
+	{
+	  fprintf(stderr, "could not read signature header\n");
+	  exit(1);
+	}
+      d.h = readhead(nfd, 0);
+      if (!d.h)
+	{
+	  fprintf(stderr, "could not read header\n");
+	  exit(1);
+	}
+    }
+  rpmMD5Init(&fullmd5);
+  rpmMD5Update(&fullmd5, rpmlead, 96);
+  rpmMD5Update(&fullmd5, sigh->intro, 16);
+  rpmMD5Update(&fullmd5, sigh->data, sigh->cnt * 16 + sigh->dcnt);
+  rpmMD5Update(&fullmd5, d.h->intro, 16);
+  rpmMD5Update(&fullmd5, d.h->data, d.h->cnt * 16 + d.h->dcnt);
+  if (rpmonly)
+    {
+      /* add new header to cpio */
+      addtocpio(&newcpio, &newcpiolen, d.h->intro, 16);
+      addtocpio(&newcpio, &newcpiolen, d.h->data, 16 * d.h->cnt + d.h->dcnt);
+    }
+  fullsize = 96 + 16 + sigh->cnt * 16 + sigh->dcnt + 16 + d.h->cnt * 16 + d.h->dcnt;
+  newbz = cfile_open(CFILE_OPEN_RD, nfd, 0, CFILE_COMP_XX, CFILE_LEN_UNLIMITED, (cfile_ctxup)rpmMD5Update, &fullmd5);
+  if (!newbz)
+    {
+      fprintf(stderr, "payload open failed\n");
+      exit(1);
+    }
+  if (cfile_detect_rsync(newbz))
+    {
+      fprintf(stderr, "detect_rsync failed\n");
+      exit(1);
+    }
+  targetcomp = newbz->comp;
+  if ((payloadflags = headstring(d.h, TAG_PAYLOADFLAGS)) != 0)
+    if (*payloadflags >= '1' && *payloadflags <= '9')
+      targetcomp = cfile_setlevel(targetcomp, *payloadflags - '0');
+  if (paycomp == CFILE_COMP_XX)
+    paycomp = targetcomp;
+  if (addblkcomp == CFILE_COMP_XX)
+    addblkcomp = targetcomp;
+
+  if (stream)
+    {
+      memset(&sd, 0, sizeof(sd));
+      sd.xnewdata = newcpio;
+      sd.xnewdatal = newcpiolen;
+      sd.newf = newbz;
+      sd.bsize = stream;
+      sd.old = xmalloc(sd.bsize);
+      sd.new = xmalloc(sd.bsize);
+      if (addblkcomp != -1)
+        sd.cfa = cfile_open(CFILE_OPEN_WR, CFILE_IO_ALLOC, &d.addblk, addblkcomp, CFILE_LEN_UNLIMITED, 0, 0);
+      sd.cfi = cfile_open(CFILE_OPEN_WR, CFILE_IO_ALLOC, &d.indata, CFILE_COMP_UN, CFILE_LEN_UNLIMITED, 0, 0);
+      oldcpio = (void *)&sd;
+      oldcpiolen = 0;
+      sd.stepd = mkdiff_step_setup(DELTAMODE_HASH | (addblkcomp == -1 ? DELTAMODE_NOADDBLK : 0));
+    }
+  else
+    {
       while ((l = newbz->read(newbz, buf, sizeof(buf))) > 0)
 	addtocpio(&newcpio, &newcpiolen, (unsigned char *)buf, l);
       if (l < 0)
@@ -809,24 +1119,17 @@ main(int argc, char **argv)
 	  fprintf(stderr, "payload read failed\n");
 	  exit(1);
 	}
-      fullsize += newbz->bytes;
-      if (newbz->close(newbz))
-	{
-	  fprintf(stderr, "junk at end of payload\n");
-	  exit(1);
-	}
-      if (strcmp(rpmname, "-") != 0)
-	close(fd);
-      fd = -1;
-      rpmMD5Final(fullmd5res, &fullmd5);
     }
 
+/***************************************************************************/
+
   if (rpmonly)
     {
       /* add old header to cpio */
       addtocpio(&oldcpio, &oldcpiolen, h->intro, 16);
       addtocpio(&oldcpio, &oldcpiolen, h->data, 16 * h->cnt + h->dcnt);
-      rpmMD5Update(&seqmd5, oldcpio, oldcpiolen);
+      rpmMD5Update(&seqmd5, h->intro, 16);
+      rpmMD5Update(&seqmd5, h->data, 16 * h->cnt + h->dcnt);
       bfd = cfile_open(CFILE_OPEN_RD, fd, 0, CFILE_COMP_XX, CFILE_LEN_UNLIMITED, (cfile_ctxup)rpmMD5Update, &seqmd5);
     }
   else if (alone)
@@ -863,12 +1166,17 @@ main(int argc, char **argv)
 	  cpiopos += sizeof(cph);
 	  if (memcmp(cph.magic, "070701", 6))
 	    {
-	      fprintf(stderr, "bad cpio archive\n");
+	      fprintf(stderr, "bad cpio archive in old deltarpm\n");
+		if (verbose >= 2)
+		  {
+		    fprintf(stderr, "cph.magic: %02x%02x %02x%02x %02x%02x\n", cph.magic[0], cph.magic[1], cph.magic[2], cph.magic[3], cph.magic[4], cph.magic[5]);
+		    fprintf(stderr, "Expected: 3037 3037 3031\n");
+		  }
 	      exit(1);
 	    }
 	  size = cpion(cph.filesize);
 	  nsize = cpion(cph.namesize);
-	  nlink = cpion(cph.nlink);
+	  nlink = cpion(cph.nlink); (void)nlink;
 	  nsize += (4 - ((nsize + 2) & 3)) & 3;
 	  if (nsize > namebufl)
 	    {
@@ -1115,7 +1423,28 @@ oaretry1:
 	}
     }
   bfd->close(bfd);
-  if (fd != -1 && strcmp(rpmname, "-") != 0)
+
+  if (stream)
+    {
+      /* finish */
+      sd.oldeof = 1;
+      addtocpio(&oldcpio, &oldcpiolen, 0, 0);
+      if (sd.cfa)
+        d.addblklen = sd.cfa->close(sd.cfa);
+      d.inlen = sd.cfi->close(sd.cfi);
+      sd.cfa = 0;
+      sd.cfi = 0;
+      instr = sd.instr;
+      sd.instr = 0;
+      instrlen = sd.instrlen;
+      sd.instrlen = 0;
+      mkdiff_step_free(sd.stepd);
+      oldcpio = 0;
+    }
+
+  /* close old rpm */
+  /* fd == -1 in "alone" mode */
+  if (fd != -1 && strcmp(argv[argc - 3 + alone], "-") != 0)
     close(fd);
 
   rpmMD5Final(seqmd5res, &seqmd5);
@@ -1129,110 +1458,41 @@ oaretry1:
   filelinktos = xfree(filelinktos);
   filenames = xfree(filenames);
   filecolors = xfree(filecolors);
-  if (!alone)
-    h = xfree(h);
+  h = xfree(h);
+
+  /* close new rpm */
+  fullsize += newbz->bytes;
+  if (newbz->close(newbz))
+    {
+      fprintf(stderr, "junk at end of payload\n");
+      exit(1);
+    }
+  if (strcmp(rpmname, "-") != 0)
+    close(nfd);
+  rpmMD5Final(fullmd5res, &fullmd5);
 
 /****************************************************************/
 
-  if (!alone)
-    {
-      /* read in new rpm */
-      if (verbose)
-	fprintf(vfp, "reading new rpm...\n");
-      rpmname = argv[argc - 2];
-      if (!strcmp(rpmname, "-"))
-	fd = 0;
-      else if ((fd = open(rpmname, O_RDONLY)) < 0)
-	{
-	  perror(rpmname);
-	  exit(1);
-	}
-      if (read(fd, rpmlead, 96) != 96 || rpmlead[0] != 0xed || rpmlead[1] != 0xab || rpmlead[2] != 0xee || rpmlead[3] != 0xdb)
-	{
-	  fprintf(stderr, "%s: not a rpm\n", rpmname);
-	  exit(1);
-	}
-      if (rpmlead[4] != 0x03 || rpmlead[0x4e] != 0 || rpmlead[0x4f] != 5)
-	{
-	  fprintf(stderr, "%s: not a v3 rpm or not new header styles\n", rpmname);
-	  exit(1);
-	}
-      sigh = readhead(fd, 1);
-      if (!sigh)
-	{
-	  fprintf(stderr, "could not read signature header\n");
-	  exit(1);
-	}
-      d.h = readhead(fd, 0);
-      if (!d.h)
-	{
-	  fprintf(stderr, "could not read header\n");
-	  exit(1);
-	}
-      rpmMD5Init(&fullmd5);
-      rpmMD5Update(&fullmd5, rpmlead, 96);
-      rpmMD5Update(&fullmd5, sigh->intro, 16);
-      rpmMD5Update(&fullmd5, sigh->data, sigh->cnt * 16 + sigh->dcnt);
-      rpmMD5Update(&fullmd5, d.h->intro, 16);
-      rpmMD5Update(&fullmd5, d.h->data, d.h->cnt * 16 + d.h->dcnt);
-      if (rpmonly)
-	{
-	  /* add new header to cpio */
-	  addtocpio(&newcpio, &newcpiolen, d.h->intro, 16);
-	  addtocpio(&newcpio, &newcpiolen, d.h->data, 16 * d.h->cnt + d.h->dcnt);
-	}
-      fullsize = 96 + 16 + sigh->cnt * 16 + sigh->dcnt + 16 + d.h->cnt * 16 + d.h->dcnt;
-      newbz = cfile_open(CFILE_OPEN_RD, fd, 0, CFILE_COMP_XX, CFILE_LEN_UNLIMITED, (cfile_ctxup)rpmMD5Update, &fullmd5);
-      if (!newbz)
-	{
-	  fprintf(stderr, "payload open failed\n");
-	  exit(1);
-	}
-      if (cfile_detect_rsync(newbz))
-	{
-	  fprintf(stderr, "detect_rsync failed\n");
-	  exit(1);
-	}
-      targetcomp = newbz->comp;
-      if ((payloadflags = headstring(d.h, TAG_PAYLOADFLAGS)) != 0)
-	if (*payloadflags >= '1' && *payloadflags <= '9')
-	  targetcomp = cfile_setlevel(targetcomp, *payloadflags - '0');
-      if (paycomp == CFILE_COMP_XX)
-	paycomp = targetcomp;
-      if (addblkcomp == CFILE_COMP_XX)
-	addblkcomp = targetcomp;
-      while ((l = newbz->read(newbz, buf, sizeof(buf))) > 0)
-	addtocpio(&newcpio, &newcpiolen, (unsigned char *)buf, l);
-      if (l < 0)
-	{
-	  fprintf(stderr, "payload read failed\n");
-	  exit(1);
-	}
-      fullsize += newbz->bytes;
-      if (newbz->close(newbz))
-	{
-	  fprintf(stderr, "junk at end of payload\n");
-	  exit(1);
-	}
-      if (strcmp(rpmname, "-") != 0)
-	close(fd);
-      rpmMD5Final(fullmd5res, &fullmd5);
-    }
-
   payformat = headstring(d.h, TAG_PAYLOADFORMAT);
   if (!payformat || strcmp(payformat, "cpio") != 0)
     {
       fprintf(stderr, "payload format is not cpio\n");
       exit(1);
     }
-  if (verbose)
-    fprintf(vfp, "creating diff...\n");
-  d.addblk = 0;
-  d.addblklen = 0;
-  mkdiff(DELTAMODE_HASH | (addblkcomp == -1 ? DELTAMODE_NOADDBLK : 0), oldcpio, oldcpiolen, newcpio, newcpiolen, &instr, &instrlen, (unsigned char **)0, (unsigned int *)0, (addblkcomp == CFILE_COMP_BZ ? &d.addblk : 0), (addblkcomp == CFILE_COMP_BZ ? &d.addblklen : 0), (unsigned char **)0, (unsigned int *)0);
+  if (!stream)
+    {
+      if (verbose)
+	fprintf(vfp, "creating diff...\n");
+      d.addblk = 0;
+      d.addblklen = 0;
+      mkdiff(DELTAMODE_HASH | (addblkcomp == -1 ? DELTAMODE_NOADDBLK : 0), oldcpio, oldcpiolen, newcpio, newcpiolen, &instr, &instrlen, (unsigned char **)0, (unsigned int *)0, (addblkcomp == CFILE_COMP_BZ ? &d.addblk : 0), (addblkcomp == CFILE_COMP_BZ ? &d.addblklen : 0), (unsigned char **)0, (unsigned int *)0);
+    }
+
+/****************************************************************/
+
   if (verbose)
     fprintf(vfp, "writing delta rpm...\n");
-  if (addblkcomp != -1 && addblkcomp != CFILE_COMP_BZ)
+  if (!stream && addblkcomp != -1 && addblkcomp != CFILE_COMP_BZ)
     createaddblock(instr, instrlen, &d, oldcpio, newcpio, addblkcomp);
   d.name = argv[argc - 1];
   d.version = 0x444c5430 + version;
@@ -1242,7 +1502,9 @@ oaretry1:
   memcpy(d.lead, rpmlead, 96);
   memcpy(d.lead + 96, sigh->intro, 16);
   memcpy(d.lead + 96 + 16, sigh->data, d.leadl - 96 - 16);
-  indatalist = convertinstr(instr, instrlen, &d, newcpio);
+  convertinstr(instr, instrlen, &d);
+  if (!stream)
+    indatalist = createindatalist(instr, instrlen, &d, newcpio);
   d.nevr = nevr;
   d.seql = 16 + (seqp + 1) / 2;
   d.seq = xmalloc(d.seql);
@@ -1274,6 +1536,7 @@ oaretry1:
   sigh = xfree(sigh);
   d.h = xfree(d.h);
   indatalist = xfree(indatalist);
+  d.indata = xfree(d.indata);
   d.in = xfree(d.in);
   d.out = xfree(d.out);
   d.inn = d.outn = 0;
@@ -1283,5 +1546,10 @@ oaretry1:
   d.leadl = 0;
   nevr = xfree(nevr);
   seq = xfree(seq);
+  sd.old = xfree(sd.old);
+  sd.new = xfree(sd.new);
+  offadjs = xfree(offadjs);
+  d.targetnevr = xfree(d.targetnevr);
   exit(0);
 }
+
